// ---------------------------------------------------------------------------
//
//  Project goals:
//
//  Should allow:
//    1. USB connectivity
//    1.1. Updating scripts and font(s) via USB
//    1.2. Updating firmware via USB (encrypted)
//    2. Script processing
//    3. Input signals' processing
//
// ---------------------------------------------------------------------------
//
// Rows               - RB0..RB4
// Columns (counters) - RC0,RC1,RC2,RC6,RC7
// Reset              - RB5
// Inputs             - RA0 (Left turn), RA1 (Stop), RA2 (Right turn), RA3 (Reverse)
//
// ---------------------------------------------------------------------------

/** I N C L U D E S **********************************************************/
#include <p18cxxx.h>
#include <usart.h>
#include "system\typedefs.h"

#include "system\usb\usb.h"

#include "io_cfg.h"             // I/O pin mapping
#include "user\user.h"
#include "user\scripts.h"

/** V A R I A B L E S ********************************************************/
byte	intState;
byte	lineNo;

byte	Delay1;		// Calibration counter. Brings delay decrements down to 0.01 sec
word	Delay2;		// 0.002 sec counter
byte	Delay3;		// Input logic base delay counter
word	CharDelay;	// 0.002 sec counter
word	pDelay1,pDelay2;	// Power management mode counters

BIT_FLAGS BitFlags;

// Display control variables

byte	scrStop;
byte	Brt;		// Brightness

// Script flow control

POINTER	CallStackAddr[16];
byte	CallStackScr[16];
byte	CallStackPos;
byte	RereadFlag;

// Script control variables

byte	ScriptNumber;
POINTER	ScriptAddr;
POINTER	ScriptLimit;
word	Scripts[32];

// Character control variables

byte	ParsedChar[32];
byte	CharColPos;
word	CharsLeft;
byte	CharLen;
byte	CharSpace;

// Input logic variables
byte	cLeft,cRight,cStop,cReverse;	// Input counters
byte	iLeft,iRight;					// Input flags
byte	pLeft,pRight;					// Number of Left/Right blinks
word	tLeft,tRight;					// Sequence timeouts
INPUT_FLAGS srcFlags,memFlags;

// Insert buffers

BIT_FIELD	vInsert;

#pragma udata usbbuf

// USB Communications

char	CmdIn[70];
byte	CmdInSz;
byte	ProcessStage;

// ---------------------------------------------------------------------------

#pragma udata scrbuf

tRowMatrix	matrix0;
byte	display0,display1,display2,display3,display4;

/** C O M M U N I C A T I O N   M E S S A G E S ******************************/	
#pragma	romdata

rom char CrLf[] = "\r\n";
rom char ErrCharStr[] = "ERR01";		// Received an incorrect character
rom char ErrOvrStr[]  = "ERR02";		// Input buffer overrun
rom char ErrCmdStr[]  = "ERR03";		// Unknown command character

rom char OkStr[] = "OK";

rom char InfoStr[] = "SLV1102a by CycleElectronics - 20070219";

// Script Area Size	XXXX (these bytes are not in the following string,
//                        they are being sent in a separate transaction)
rom char IDStr[] = "SL00110002";

/** D E C L A R A T I O N S **************************************************/
#pragma code

int AddToBuf(byte Tetr) {
	byte Addr;
	Addr = CmdInSz>>1;
	if((CmdInSz&1)==0) {
		CmdIn[Addr] = (Tetr<<4);
		CmdInSz++;
	}
	else {
		CmdIn[Addr] |= Tetr;
		if((++CmdInSz)>=140)	return 0;
	}
	return 1;
}

// ---------------------------------------------------------------------------

byte ByteToHex(byte b) {
	b &= 0x0f;
	if(b > 9)
		return (b+0x37);
	else
		return (b+0x30);
}

// ---------------------------------------------------------------------------

void Prg16(word Addr, byte Offset) {
	word ta;
	byte ii;
	
	ta = Addr+((word)Offset);
	for(ii=0;ii<16;ii++) {
		(*(rom byte*)ta) = CmdIn[ii+Offset+2];
		ta++;
	}
	
	TBLPTR = (Addr+((word)Offset));
	EECON1bits.CFGS = 0;			// access Flash program memory
	EECON1bits.WREN = 1;			// enable write to memory
	INTCONbits.GIE  = 0;			// disable interrupts
	EECON2          = 0x55;
	EECON2          = 0xaa;
	EECON1bits.WR   = 1;			// start program (CPU stall)
	INTCONbits.GIE  = 1;			// re-enable interrupts
	EECON1bits.WREN = 0;			// disable write to memory
}

// ---------------------------------------------------------------------------

void ReinitUSB(void) {
	ProcessStage = STAGE_START;
	CmdInSz = 0;
}

// ---------------------------------------------------------------------------

void ProcessUsbCommand(void) {
	byte Tmp;
	char ss[2];
	word tw;

	switch(ProcessStage) {
		case STAGE_START:
			if(mCDCUsartRxIsBusy()) return;
			if(getsUSBUSART(&Tmp,1)) {
				if(Tmp==0x0d) {				// Run the command
					if((CmdInSz>1)&&(CmdIn[0]=='I')) {
						ProcessStage = STAGE_ID;
					}
					else if((CmdInSz>3)&&(CmdIn[0]=='R')) {
						tw = ((word)CharSizeTbl)+(((word)CmdIn[1])*((word)32));
						for(Tmp=0;Tmp<64;Tmp+=2) {
					        CmdIn[Tmp] = ByteToHex((*((rom byte*)tw))>>4);
					        CmdIn[Tmp+1] = ByteToHex(*((rom byte*)tw));
					        tw++;
						}
						pCDCSrc.bRam = (byte*) CmdIn;
						cdc_tx_len = 64;
						cdc_mem_type = _RAM;
						cdc_trf_state = CDC_TX_BUSY;
				        ProcessStage = STAGE_SEND_CRLF;
					}
					else if((CmdInSz>131)&&(CmdIn[0]=='W')) {
						tw = ((word)CharSizeTbl)+(((word)CmdIn[1])*((word)64));

						TBLPTR = tw;
						EECON1bits.CFGS = 0;
						EECON1bits.WREN = 1;
						EECON1bits.FREE = 1;
						INTCONbits.GIE  = 0;
						EECON2 = 0x55;
						EECON2 = 0xaa;
						EECON1bits.WR   = 1;
						INTCONbits.GIE  = 1;
						
						Prg16(tw,0);
						Prg16(tw,16);
						Prg16(tw,32);
						Prg16(tw,48);

				        ProcessStage = STAGE_SEND_CNMF;
				        RereadFlag = 1;
					}
					else
						ProcessStage = STAGE_ERR_CMD;
				}
				else if(Tmp=='U') {			// Resync
					ReinitUSB();
				}
				else if(Tmp=='?') {			// Send InfoStr
					ProcessStage = STAGE_SEND_INFO;
				}
				else if((Tmp>='0')&&(Tmp<='9')) {
					if(!AddToBuf(Tmp&0x0f)) ProcessStage = STAGE_ERR_OVR;
				}
				else if((Tmp>='A')&&(Tmp<='F')) {
					if(!AddToBuf(Tmp-0x37)) ProcessStage = STAGE_ERR_OVR;
				}
				else if((Tmp>='a')&&(Tmp<='f')) {
					if(!AddToBuf(Tmp-0x57)) ProcessStage = STAGE_ERR_OVR;
				}
				else {
					ProcessStage = STAGE_ERR_CHAR;
				}
			}
			break;
			
		case STAGE_ID:
	        if(!mUSBUSARTIsTxTrfReady()) return;

			pCDCSrc.bRom = (rom byte*) IDStr;
			cdc_tx_len = 10;
			cdc_mem_type = _ROM;
			cdc_trf_state = CDC_TX_BUSY;

	        ProcessStage = STAGE_MEM;
	        break;
	        
		case STAGE_MEM:
	        if(!mUSBUSARTIsTxTrfReady()) return;
	        tw = TopMem-(word)CharSizeTbl;
	        CmdIn[0] = ByteToHex(tw>>12);
	        CmdIn[1] = ByteToHex(tw>>8);
	        CmdIn[2] = ByteToHex(tw>>4);
	        CmdIn[3] = ByteToHex(tw);
			pCDCSrc.bRam = (byte*) CmdIn;
			cdc_tx_len = 4;
			cdc_mem_type = _RAM;
			cdc_trf_state = CDC_TX_BUSY;
	        ProcessStage = STAGE_SEND_CRLF;
	        break;

		case STAGE_SEND_INFO:
	        if(!mUSBUSARTIsTxTrfReady()) return;
	        putrsUSBUSART(InfoStr);
	        ProcessStage = STAGE_SEND_CRLF;
	        break;
	        
		case STAGE_SEND_CNMF:
	        if(!mUSBUSARTIsTxTrfReady()) return;
	        putrsUSBUSART(OkStr);
	        ProcessStage = STAGE_SEND_CRLF;
	        break;

		case STAGE_SEND_CRLF:
	        if(!mUSBUSARTIsTxTrfReady()) return;
	        putrsUSBUSART(CrLf);
	        ReinitUSB();
	        break;
	        
		case STAGE_ERR_CHAR:
	        if(!mUSBUSARTIsTxTrfReady()) return;
	        putrsUSBUSART(ErrCharStr);
	        ProcessStage = STAGE_SEND_CRLF;
	        break;
	        
		case STAGE_ERR_OVR:
	        if(!mUSBUSARTIsTxTrfReady()) return;
	        putrsUSBUSART(ErrOvrStr);
	        ProcessStage = STAGE_SEND_CRLF;
	        break;
	        
		case STAGE_ERR_CMD:
	        if(!mUSBUSARTIsTxTrfReady()) return;
	        putrsUSBUSART(ErrCmdStr);
	        ProcessStage = STAGE_SEND_CRLF;
	        break;
	}
}

// -----------------------------------------------------------------
// -----------------------------------------------------------------
// -----------------------------------------------------------------

void UserInit(void)
{
	OSCCON	= 0b00000000;
	OSCTUNE = 0b00000000;
	
	// Initializing interrupt controller

	INTCON  = 0;

	INTCON  = 0b00100000;		// Disable all interrupts
	INTCON2	= 0x00;				// Configure interrupts so we will
	INTCON3	= 0x00;				// process only TIMER0 interrupts
	PIR1    = 0x00;
	PIR2    = 0x00;
	PIE1    = 0x00;
	PIE2    = 0x00;
	IPR1    = 0x00;
	IPR2    = 0x00;
	RCONbits.IPEN = 0;

	ADCON0  = 0;
	ADCON1  = 0x0f;
	ADCON2  = 0;
	CMCON   = 0x07;
	CCP1CON = 0;
	CCP2CON = 0;
	SSPCON1 = 0;
	TXSTA   = 0;
	RCSTA   = 0;

	// Initializing PORTA
	LATA    = 0;
	TRISA   = 0xff;

	// Initializing PORTB
	#ifdef V1
	LATB    = 0b11100000;
	#endif
	#ifdef V2
	LATB    = 0b11100000;
	#endif
	#ifdef V3
	LATB    = 0b11000000;
	#endif
	// V3
	TRISB   = 0xc0;

	// Initializing PORTC
	LATC    = 0;
	TRISC   = 0x0;

	// Initializing TIMER0
	//   Device clock = PLL/2 = 48MHz
	//   To make the period of the TIMER1 interrupt to be equal to 200 mks (0.0002 sec)
	//   (5000 interrupts per second) TIMER0 VALUE should be compliment to 150 (0x96)
	//   with a prescaler set to 16.
	T0CON	= 0b11000001;
	TMR0L	= TIMER0_VALUE;

	// We need to initialize the device only if it is Power On Reset (POR)
	// Otherwise, we just continue the script execution

	if(!RCONbits.POR) PorReset();

	//  ------------	Enabling interrupts
		
	RCONbits.POR      = 1;		// To catch non-PowerOn resets
	RCONbits.IPEN     = 1;		// Allow priority levels
	IPR2bits.USBIP    = 1;		// Assign high priority to the USB interrupt
	INTCONbits.GIE    = 1;		// Enable interrupts and wait for one
	INTCONbits.PEIE   = 1;		// Enable periferal interrupts and wait for one
	INTCONbits.TMR0IE = 1;		// Enable Timer0 interrupt, keep it in low priority
}//end UserInit

// ---------------------------------------------------------------------------

void PorReset(void) {
	byte ii,sn;
	word initDelay;
	POINTER	ScriptStart;
	
	RereadFlag = 0;
	
	// Calculating the begginning of the script table
	ScriptStart.bRom = &(CharTable[0]);
	for(ii=0;ii<128;ii++)ScriptStart._word+=CharSizeTbl[ii];

	// Filling in the script table
	Scripts[0]=ScriptStart._word;
	sn = 1;
	do {
		if((*(ScriptStart.bRom++))==0xaa) {
			if((*(ScriptStart.bRom++))==0x55) {
				Scripts[sn++] = ScriptStart._word;
			}
		}
	} while((ScriptStart._word<TopMem)&(sn<32));
	
	InitVar();
	
	// Initial delay
	Delay2 = 0xffff;
	while(--Delay2);
	
	if(PORTAbits.RA1)
		ScriptNumber = 0x01;
	else
		ScriptNumber = 0x10;
	CalcScriptAddr();
	
	for(initDelay=0;initDelay<0xffff;initDelay++);
	for(initDelay=0;initDelay<0xffff;initDelay++);
}

// ---------------------------------------------------------------------------

void CalcScriptAddr(void) {
    byte ii;

	if(ScriptNumber>31)ScriptNumber=0;
	ScriptAddr._word = Scripts[ScriptNumber];
	ScriptLimit._word = TopMem;
	for(ii=0;ii<32;ii++)
		if((ii!=ScriptNumber)&&(Scripts[ii]>ScriptAddr._word)&&(Scripts[ii]<ScriptLimit._word))
			ScriptLimit._word = Scripts[ii];
	scrStop = 0;
}

// ---------------------------------------------------------------------------

void InitVar(void) {
	byte ii,jj;
	
	intState	= 0;
	Delay1		= 0;
	Delay2		= 0;
	Delay3		= 0;
	CharDelay	= 40;
	lineNo		= 0;
	scrStop		= 0;
	Brt			= 7;

	pDelay1 = pDelay2 = 0;
	
	BitFlags.InsertBlanks = 1;
    BitFlags.CharDir = 0;
    BitFlags.CharMirroring = 0;

	CharColPos	= 0;
	CharsLeft	= 0;
	CharLen		= 0;
	CharSpace	= 0;
	
	CallStackPos = 0;

	for(ii=0;ii<5;ii++) for(jj=0;jj<5;jj++) matrix0[ii][jj] = 0;
	
	display0 = 0;
	display1 = 0;
	display2 = 0;
	display3 = 0;
	display4 = 0;
	
	iLeft = iRight = 0;
	cLeft = cRight = cStop = cReverse = 0;
	pLeft = pRight = 0;
	tLeft = tRight = 0;
	srcFlags._byte = 0;
	memFlags._byte = 0;

	CmdInSz                = 0;				// Must be cleared if USB gets disconnected!!!
	ProcessStage           = STAGE_START;	// ...
}

// ---------------------------------------------------------------------------

void InsertFromLeft(BIT_FIELD fld) {
	ShiftMatrixRight(0,5);
	matrix0[0][0] = (matrix0[0][0] & 0b01111110); if(fld.Bit0) matrix0[0][0] |= 0b00000001;
	matrix0[1][0] = (matrix0[1][0] & 0b01111110); if(fld.Bit1) matrix0[1][0] |= 0b00000001;
	matrix0[2][0] = (matrix0[2][0] & 0b01111110); if(fld.Bit2) matrix0[2][0] |= 0b00000001;
	matrix0[3][0] = (matrix0[3][0] & 0b01111110); if(fld.Bit3) matrix0[3][0] |= 0b00000001;
	matrix0[4][0] = (matrix0[4][0] & 0b01111110); if(fld.Bit4) matrix0[4][0] |= 0b00000001;
}

// ---------------------------------------------------------------------------

void InsertFromRight(BIT_FIELD fld) {
	ShiftMatrixLeft(0,5);
	matrix0[0][4] = (matrix0[0][4] & 0b00111111);  if(fld.Bit0) matrix0[0][4] |= 0b01000000;
	matrix0[1][4] = (matrix0[1][4] & 0b00111111);  if(fld.Bit1) matrix0[1][4] |= 0b01000000;
	matrix0[2][4] = (matrix0[2][4] & 0b00111111);  if(fld.Bit2) matrix0[2][4] |= 0b01000000;
	matrix0[3][4] = (matrix0[3][4] & 0b00111111);  if(fld.Bit3) matrix0[3][4] |= 0b01000000;
	matrix0[4][4] = (matrix0[4][4] & 0b00111111);  if(fld.Bit4) matrix0[4][4] |= 0b01000000;
}

// ---------------------------------------------------------------------------

void ParseChar(byte Character) {
	POINTER CharPosBuf;
	byte ii;
	
	CharPosBuf.bRom = &(CharTable[0]);
	for(ii=0;ii<Character;ii++)
		CharPosBuf._word = CharPosBuf._word+CharSizeTbl[ii];
	CharLen = CharSizeTbl[Character];
	for(ii=0;ii<CharLen;ii++)
		ParsedChar[ii] = *(CharPosBuf.bRom++);
}

// ---------------------------------------------------------------------------

int GetChar(byte GetCurrent) {
	// Loads next char from the script stream, parses it, fills in all the
	// related variables.
	// Returns 0 if there was script sizing error.
	
	if(!CharsLeft) {
		CharLen = 0;
		return 1;
	}
	if(!GetCurrent)if(IncScriptAddr())return 0;
	if((*(ScriptAddr.bRom))==0) {
		CharsLeft = 0;
		CharLen = 0;
		return 0;
	}
	ParseChar(*(ScriptAddr.bRom)); // Fills in CharLen
	if(BitFlags.CharMirroring)
		CharColPos = CharLen;
	else
		CharColPos = 0;
	return 1;
}

// ---------------------------------------------------------------------------

void ShiftMatrixLeft(byte Strt, byte Stp) {
	byte i;
	for(i=Strt;i<Stp;i++) {
		if(matrix0[i][4]&1) matrix0[i][3] |= 0b10000000; else matrix0[i][3] &= 0b01111111;
		if(matrix0[i][3]&1) matrix0[i][2] |= 0b10000000; else matrix0[i][2] &= 0b01111111;
		if(matrix0[i][2]&1) matrix0[i][1] |= 0b10000000; else matrix0[i][1] &= 0b01111111;
		if(matrix0[i][1]&1) matrix0[i][0] |= 0b10000000; else matrix0[i][0] &= 0b01111111;
		matrix0[i][0] >>= 1;
		matrix0[i][1] >>= 1;
		matrix0[i][2] >>= 1;
		matrix0[i][3] >>= 1;
		matrix0[i][4] >>= 1;
		matrix0[i][4] &= 0b00111111;
		matrix0[i][3] &= 0b01111111;
		matrix0[i][2] &= 0b01111111;
		matrix0[i][1] &= 0b01111111;
		matrix0[i][0] &= 0b01111111;
	}
}

// ---------------------------------------------------------------------------

void ShiftMatrixRight(byte Strt, byte Stp) {
	byte i;
	for(i=Strt;i<Stp;i++) {
		matrix0[i][0] <<= 1;
		matrix0[i][1] <<= 1;
		matrix0[i][2] <<= 1;
		matrix0[i][3] <<= 1;
		matrix0[i][4] <<= 1;
		if(matrix0[i][0]&0b10000000) matrix0[i][1] |= 0b00000001; else matrix0[i][1] &=0b11111110;
		if(matrix0[i][1]&0b10000000) matrix0[i][2] |= 0b00000001; else matrix0[i][2] &=0b11111110;
		if(matrix0[i][2]&0b10000000) matrix0[i][3] |= 0b00000001; else matrix0[i][3] &=0b11111110;
		if(matrix0[i][3]&0b10000000) matrix0[i][4] |= 0b00000001; else matrix0[i][4] &=0b11111110;
		matrix0[i][0] &= 0b01111110;
		matrix0[i][1] &= 0b01111111;
		matrix0[i][2] &= 0b01111111;
		matrix0[i][3] &= 0b01111111;
		matrix0[i][4] &= 0b01111111;
	}
}

// ---------------------------------------------------------------------------

void ReadInputs(void) {
	if(!Delay3) {
		Delay3 = C_DELAY;
		
		// Check if the "idle" timeout has expired
		pDelay1+=1;
		if(pDelay1==0xffff) {
			// We are here approximately every 130 sec
			pDelay1=0;
			if(pDelay2<((word)0x00ff))pDelay2++;
			if(pDelay2>=offTimeout) {
				// Entering power saving mode
				T0CONbits.TMR0ON = 0;
				#ifdef V1
				LATB = 0b11100000;
				#endif
				#ifdef V2
				LATB = 0b11100000;
				#endif
				#ifdef V3
				LATB = 0b11000000;
				#endif
				return;
			}
		}

		// Accessory power input
		srcFlags.AccPwr = PORTAbits.RA5;

		// Left turn
		if(PORTAbits.RA0) {
			if(cLeft<DETECT_LVL)
				cLeft++;
			else
				iLeft = 1;
		}
		else {
			if(cLeft) {
				if(!(--cLeft))pLeft++;
			}
			else
				iLeft = 0;
		}
		
		// Break
		if(PORTAbits.RA1) {
			if(cStop<DETECT_LVL)
				cStop++;
			else
				srcFlags.Break = 1;
		}
		else {
			if(cStop)
				cStop--;
			else
				srcFlags.Break = 0;
		}
		
		// Right turn
		if(PORTAbits.RA2) {
			if(cRight<DETECT_LVL)
				cRight++;
			else
				iRight = 1;
		}
		else {
			if(cRight) {
				if(!(--cRight))pRight++;
			}
			else
				iRight = 0;
		}
		
		// Reverse
		if(PORTAbits.RA3) {
			if(cReverse<DETECT_LVL)
				cReverse++;
			else
				srcFlags.Reverse = 1;
		}
		else {
			if(cReverse)
				cReverse--;
			else
				srcFlags.Reverse = 0;
		}
	}
}

// ---------------------------------------------------------------------------

void DetectSequences(void) {
	if(iLeft)
		tLeft = SEQ_DETECT_LVL;
	else {
		if(tLeft)
			tLeft--;
	}
	if(iRight)
		tRight = SEQ_DETECT_LVL;
	else {
		if(tRight)
			tRight--;
	}
	// Set the flags
	if(!tLeft)
		srcFlags.LeftTurn = 0;
	if(!tRight)
		srcFlags.RightTurn = 0;
	if(tLeft&!tRight)
		srcFlags.LeftTurn = 1;
	if(tRight&!tLeft)
		srcFlags.RightTurn = 1;
	if(tLeft&tRight&(pLeft>3)&(pRight>3))
		srcFlags.Emergency = 1;
	else
		srcFlags.Emergency = 0;
}

// ---------------------------------------------------------------------------

void CheckEvents(void) {
	if(srcFlags._byte!=memFlags._byte) {
		memFlags._byte=srcFlags._byte;
		pDelay1 = pDelay2 = 0;
		pLeft = pRight = 0;
		if(!memFlags._byte)
			ScriptNumber = 0;
		else if(memFlags.Emergency)
			ScriptNumber = SCR_EMERGENCY;
		else if(memFlags.Break)
			ScriptNumber = SCR_BREAK;
		else if(memFlags.LeftTurn)
			ScriptNumber = SCR_LEFT;
		else if(memFlags.RightTurn)
			ScriptNumber = SCR_RIGHT;
		else if(memFlags.Reverse)
			ScriptNumber = SCR_REVERSE;
		else if(memFlags.AccPwr)
			ScriptNumber = SCR_COMPL;
		else
			ScriptNumber = 0;
		CalcScriptAddr();
		CharsLeft = 0;
		CharLen = 0;
		Delay1 = 0;
		Delay2 = 0;
		CallStackPos = 0;
	}
}

// ---------------------------------------------------------------------------

void ProcessIO(void)
{
	BIT_FIELD fld;
	byte ii;

	// Power saving mode loop
	while(pDelay2>=offTimeout) {
		if(PORTA&0b00011111) {
			T0CONbits.TMR0ON=1;
			pDelay1 = pDelay2 = 0;
			pLeft = pRight = 0;
			ScriptNumber = 0x10;
			CalcScriptAddr();
			CharsLeft = 0;
			CharLen = 0;
			Delay1 = 0;
			Delay2 = 0;
			CallStackPos = 0;
		}
	}

	if(usb_bus_sense == 1) {
		if((usb_device_state >= CONFIGURED_STATE)&&(UCONbits.SUSPND==0)) {
			ProcessUsbCommand();
			return;
		}
	}
	
	if(RereadFlag) {
		PorReset();
		CalcScriptAddr();
	}
	
	// Read Inputs
	ReadInputs();
	DetectSequences();
	CheckEvents();
	
	if(scrStop||Delay1) return;
	
	if(Delay2) {
		Delay2--;
		Delay1 = C_DELAY;
		return;
	}

	// Process any pending character columns
	if(CharLen) {
		if(BitFlags.CharMirroring)
			fld._byte = ParsedChar[--CharColPos];
		else
			fld._byte = ParsedChar[CharColPos++];
		if(BitFlags.CharDir)
			InsertFromLeft(fld);
		else
			InsertFromRight(fld);
		Delay2 = CharDelay;
		Delay1 = C_DELAY;
		if((--CharLen)==0) {
			CharSpace = BitFlags.InsertBlanks;
			if(!(--CharsLeft)) {
				if(IncScriptAddr())return;
			}
		}
		return;
	}
	else {
		if(CharSpace) {
			CharSpace--;
			fld._byte = 0;
			if(BitFlags.CharDir)
				InsertFromLeft(fld);
			else
				InsertFromRight(fld);
			Delay2 = CharDelay;
			Delay1 = C_DELAY;
			return;
		}
		else {
			// Next character?
			if(CharsLeft) {
				GetChar(0);
				return;
			}
		}
	}

	// Processing script commands
	if(ProcessCommand(*(ScriptAddr.bRom))) IncScriptAddr();

}//end ProcessIO

// ---------------------------------------------------------------------------

int ProcessCommand(byte Command) {
	byte ii,jj;
	word ww;
	
	switch(Command) {
		case 0b00000000:			// End of script
			if(CallStackPos) {
				ScriptNumber = CallStackScr[--CallStackPos];
				CalcScriptAddr();
				ScriptAddr._word = CallStackAddr[CallStackPos]._word;
			}
			else {
				CalcScriptAddr();
			}
			CharsLeft = 0;
			CharLen = 0;
			CharSpace = 0;
			return 0;
		case 0b00000001:			// Clear all
			for(ii=0;ii<5;ii++) for(jj=0;jj<5;jj++) matrix0[ii][jj] = 0;
			break;
		case 0b00000010:			// Fill all with a background intensity
			for(ii=0;ii<5;ii++) for(jj=0;jj<5;jj++) matrix0[ii][jj] = 0b01111111;
			break;
		case 0b00000011:			// Fill all with a foreground intensity
			for(ii=0;ii<5;ii++) for(jj=0;jj<5;jj++) matrix0[ii][jj] = 0b01111111;
			break;
		case 0b00000100:			// Insert From Right
			InsertFromRight(vInsert);
			break;
		case 0b00000101:			// Insert From Left
			InsertFromLeft(vInsert);
			break;
		case 0b11000000:			// Fill in the vertical insert buffer (next byte contains a value)
			if(IncScriptAddr())return 0;
			vInsert._byte = (*(ScriptAddr.bRom));
			break;
		case 0b11000001:			// Function call (next byte is a script number)
			if(IncScriptAddr())return 0;
			if(CallStackPos>15)return 1;
			CallStackScr[CallStackPos] = ScriptNumber;
			ScriptNumber = (*(ScriptAddr.bRom));
			IncScriptAddr();
			CallStackAddr[CallStackPos++]._word = ScriptAddr._word;
			CalcScriptAddr();
			return 0;
		case 0b11000010:			// GoTo (next byte is a script number)
			if(IncScriptAddr())return 0;
			ScriptNumber = (*(ScriptAddr.bRom));
			CalcScriptAddr();
			return 0;
		case 0b11000011:			// Stop execution
			if(CallStackPos) {
				ScriptNumber = CallStackScr[--CallStackPos];
				CalcScriptAddr();
				ScriptAddr._word = CallStackAddr[CallStackPos]._word;
				CharsLeft = 0;
				CharLen = 0;
				CharSpace = 0;
				return 0;
			}
			else {
				scrStop = 1;
			}
			break;
		case 0b11000100:			// Insert from top centered
			break;
		case 0b11000101:			// Insert from top aligned to the left
			break;
		case 0b11000110:			// Insert from top aligned to the right
			break;
		case 0b11000111:			// Insert from bottom centered
			break;
		case 0b11001000:			// Insert from bottom aligned to the left
			break;
		case 0b11001001:			// Insert from bottom aligned to the right
			break;
		case 0b11001010:			// Clear the horizontal insert buffer
			break;
		case 0b11001011:			// Scroll a line. Next byte is a number of characters to follow
			if(IncScriptAddr())return 0;
			if(CharsLeft=(*(ScriptAddr.bRom))) {
				GetChar(0);
				return 0;
			}
			break;
		case 0b11001100:			// Delay. Next byte is a number of 0.01 sec intervals
			if(IncScriptAddr())return 0;
			Delay2 = ((word)(*(ScriptAddr.bRom)));
			Delay1 = C_DELAY;
			break;
		case 0b11001101:			// Set the character output delay. Next byte is a number of 0.01 sec intervals
			if(IncScriptAddr())return 0;
			CharDelay = ((word)(*(ScriptAddr.bRom)));
			break;
		case 0b11001110:			// Delay. Next byte is a number of 0.1 sec intervals
			if(IncScriptAddr())return 0;
			ww = ((word)(*(ScriptAddr.bRom)));
			Delay2 = ww*((word)10);
			Delay1 = C_DELAY;
			break;
		case 0b11001111:			// Extended command set (for devices with large memory)
			break;
		default:
			if((Command&0b11111100)==0b01000000) {		// 010000xx - Set character spacing. xx = 0-3 empty columns between characters
				BitFlags.InsertBlanks = (Command&0b00000011);
			}
			else if((Command&0b11111110)==0b01000100) {	// 0100010x - Set the direction for scrolling characters: 0 - right to left (default)
				BitFlags.CharDir = (Command&0b00000001);
			}
			else if((Command&0b11111110)==0b01000110) {	// 0100011x - Char mirroring. 0 - straight, 1 - mirrored.
				BitFlags.CharMirroring = (Command&0b00000001);
			}
			else if((Command&0b11111000)==0b01100000) {	// 01100xxx - Set background brightness
				Brt = (Command&0b00000111);
			}
			else if((Command&0b11111000)==0b01101000) {	// 01101xxx - Set foreground brightness
				Brt = (Command&0b00000111);
			}
			else if((Command&0b11111000)==0b01110000) {	// 01110xxx - Scroll a row xxx left
				if((ii=(Command&0b00000111))<5) ShiftMatrixLeft(ii,ii+1);
			}
			else if((Command&0b11111000)==0b01111000) {	// 01111xxx - Scroll a row xxx right
				if((ii=(Command&0b00000111))<5) ShiftMatrixRight(ii,ii+1);
			}
			else if((Command&0b11110000)==0b11010000) {	// 1101xxxx - Scroll the line, defined in script # 1xxxx (top half only)
				if(CallStackPos>15)return 1;
				ii = (Command&0b00011111);
				CallStackScr[CallStackPos] = ScriptNumber;
				IncScriptAddr();
				CallStackAddr[CallStackPos++]._word = ScriptAddr._word;
				ScriptNumber = ii;
				CalcScriptAddr();
				CharsLeft = 0x0fff;
				GetChar(1);
				return 0;
			}
//   111byyyy - Set a pixel to a "b" value next byte is xxxxxxxx
//   10xxxxxx - Fill in xxxxxx bits of the horizontal insert buffer
	}
	return 1;
}

// ---------------------------------------------------------------------------

int IncScriptAddr(void) {
	if((++ScriptAddr._word)>=ScriptLimit._word) {
		// We have reached the end of script
		if(CallStackPos) {
			ScriptNumber = CallStackScr[--CallStackPos];
			CalcScriptAddr();
			ScriptAddr._word = CallStackAddr[CallStackPos]._word;
		}
		else {
			CalcScriptAddr();
		}
		CharsLeft = 0;
		CharLen = 0;
		CharSpace = 0;
		return 1;
	}
	else
	  return 0;
}

/** T I M E R   I N T E R R U P T   H A N D L E R ****************************/

void ShowARow(void) {
	switch(lineNo) {
		case 0:
			LATBbits.LATB0 = 1;
			break;
		case 1:
			LATBbits.LATB1 = 1;
			break;
		case 2:
			LATBbits.LATB2 = 1;
			break;
		case 3:
			LATBbits.LATB3 = 1;
			break;
		case 4:
			LATBbits.LATB4 = 1;
			break;
	}
}

// ---------------------------------------------------------------------------

#pragma interruptlow low_isr
void low_isr(void) {
	byte ii;

	TMR0L	= TIMER0_VALUE;

	// Adjusting delay counters
	if(Delay1) Delay1--;
	if(Delay3) Delay3--;
	
	if((++intState) >= 64){
		intState = 0;
		#ifdef V1
		LATB = 0b11000000;				// Turn off the display
		#endif
		#ifdef V2
		LATB = 0b11000000;				// Turn off the display
		#endif
		#ifdef V3
		LATB = 0b11100000;				// Turn off the display
		#endif
		if((++lineNo)==5) lineNo = 0;
		display0 = matrix0[lineNo][0];
		display1 = matrix0[lineNo][1];
		display2 = matrix0[lineNo][2];
		display3 = matrix0[lineNo][3];
		display4 = matrix0[lineNo][4];
		#ifdef V1
		LATBbits.LATB5 = 1;
		#endif
		#ifdef V2
		LATBbits.LATB5 = 1;
		#endif
		#ifdef V3
		LATBbits.LATB5 = 0;
		#endif
		while(display0|display1|display2|display3|display4) {
			if(display0) {
				LATCbits.LATC0 = 1;
				display0--;
			}
			if(display1) {
				LATCbits.LATC1 = 1;
				display1--;
			}
			if(display2) {
				LATCbits.LATC2 = 1;
				display2--;
			}
			if(display3) {
				LATCbits.LATC6 = 1;
				display3--;
			}
			if(display4) {
				LATCbits.LATC7 = 1;
				display4--;
			}
			LATC = 0;
		}
	}
	// Main brightness selector
	switch(intState) {
		case T_BR_7:
			if(Brt==7) ShowARow();
			break;
		case T_BR_6:
			if(Brt==6) ShowARow();
			break;
		case T_BR_5:
			if(Brt==5) ShowARow();
			break;
		case T_BR_4:
			if(Brt==4) ShowARow();
			break;
		case T_BR_3:
			if(Brt==3) ShowARow();
			break;
		case T_BR_2:
			if(Brt==2) ShowARow();
			break;
		case T_BR_1:
			if(Brt==1) ShowARow();
			break;
	}
	INTCONbits.TMR0IF = 0;
}

#pragma code

/** EOF user.c ***************************************************************/
